package com.maileon.api.blacklists;

import java.util.List;

import com.maileon.api.AbstractMaileonService;
import com.maileon.api.MaileonConfiguration;
import com.maileon.api.MaileonException;
import com.maileon.api.ResponseWrapper;

/**
 * The <code>MaileonBlacklistsService</code> client sends operational requests to the <code>BlacklistsResource</code>.
 *
 */
public class MaileonBlacklistsService extends AbstractMaileonService {

    // The name identifier of the underlying service.
    public static final String SERVICE = "MAILEON BLACKLISTS SERVICE";

    /**
     * Constructs a <code>MaileonBlacklistsService</code>.
     *
     * @param config Maileon API-Configuration.
     */
    public MaileonBlacklistsService(MaileonConfiguration config) {
        super(config, SERVICE);
    }

    /**
     * Gives a list of all defined blacklists in the account. See <a href="http://dev.maileon.com/api/rest-api-1-0/blacklists/get-blacklists/?lang=en">Maileon API documentation</a>.
     *
     * @return list of {@link Blacklist}
     * @throws MaileonException
     */
    public List<Blacklist> getBlacklists() throws MaileonException {
        ResponseWrapper response = get("blacklists");
        if (response.getStatusCode() == 200) {
            return BlacklistsXmlSerializer.deserialize(response.getEntityAsXml());
        } else {
            throw new MaileonException(response.getStatusCode(), response.getStatus().getReasonPhrase());
        }
    }

    /**
     * Gives a account-specific blacklist according to an id. See <a href="http://dev.maileon.com/api/rest-api-1-0/blacklists/get-blacklists/?lang=en">Maileon API documentation</a>.
     *
     * @param id of the required blacklist
     * @return a single {@link Blacklist} specified by the given id
     * @throws MaileonException
     */
    public Blacklist getBlacklist(long id) throws MaileonException {
        ResponseWrapper response = get("blacklists/" + Long.toString(id));
        if (response.getStatusCode() == 200) {
            return BlacklistXmlSerializer.deserialize(response.getEntityAsXml());
        } else {
            throw new MaileonException(response.getStatusCode(), response.getReasonPhrase());
        }
    }

    /**
     * Adds entries to an existing blacklist. See <a href="http://dev.maileon.com/api/rest-api-1-0/blacklists/add-entries-to-blacklist/?lang=en">Maileon API documentation</a>.
     *
     * @param id of the blacklist to change
     * @param importName name-identifier displayed in the user interface, can be null for the default name
     * @param entries list of entries to add to the underlying blacklist
     * @throws MaileonException
     */
    public void addBlacklistEntries(long id, String importName, List<String> entries) throws MaileonException {
        AddEntriesAction action = new AddEntriesAction();
        action.setImportName(importName);
        action.setEntries(entries);
        ResponseWrapper response = post("blacklists/" + id + "/actions", AddEntriesActionXmlSerializer.serialize(action).asXML());
        if (response.getStatusCode() != 201) {
            throw new MaileonException(response.getStatusCode(), response.getReasonPhrase());
        }
    }

}
