/**
 * @(#)MaileonReportsService.java
 */
package com.maileon.api.reports;

import com.maileon.api.AbstractMaileonService;
import com.maileon.api.MaileonBadRequestException;
import com.maileon.api.MaileonClientException;
import com.maileon.api.MaileonConfiguration;
import com.maileon.api.MaileonException;
import com.maileon.api.Page;
import com.maileon.api.ResponseWrapper;
import com.maileon.api.contacts.StandardContactField;
import org.dom4j.Element;

import java.util.ArrayList;
import java.util.List;

/**
 * The <code>MaileonReportsService</code> client send operational requests to the <code>ReportsResource</code>.
 *
 */
public class MaileonReportsService extends AbstractMaileonService {

    public static final String SERVICE = "MAILEON REPORTS";

    /**
     * Instantiates a new maileon reports service.
     *
     * @param config the configuration of maileon API
     */
    public MaileonReportsService(MaileonConfiguration config) {
        super(config, SERVICE);
    }

    /**
     * Returns a page of opens.
     * <p>
     * <a href="http://dev.maileon.com/api/rest-api-1-0/reports/get-opens/?lang=en">Documentation website</a></p>
     *
     * @param fromDate if provided, only the opens after the given date will be returned
     * @param toDate if provided, only the opens before the given date will be returned
     * @param mailingIdFilter multivalued parameter to filter the opens by mailings
     * @param embedEmailClientInfos ff the set to <code>true</code>, available email client details will be appended to each open
     * @param excludeAnonymousOpens if <code>true</code>, the opens that cannot be mapped to contacts because of permission types or because of unsubscriptions are omitted in the
     * results
     * @param standardFields multivalued parameter to query the value of a standard contact field
     * @param customFields multivalued parameter to query the value of a custom contact field
     * @param pageIndex the index of the result page starting from 1
     * @param pageSize the number of {@link Open} in the resulted page
     * @return page containing the corresponding opens
     * @throws MaileonException
     */
    public Page<Open> getOpens(Long fromDate, Long toDate, List<Long> mailingIdFilter, boolean embedEmailClientInfos,
            boolean excludeAnonymousOpens, List<StandardContactField> standardFields,
            List<String> customFields, int pageIndex, int pageSize) throws MaileonException {
        return getOpens(fromDate, toDate, mailingIdFilter, null, null, null, null, null, null, embedEmailClientInfos, excludeAnonymousOpens,
                standardFields, customFields, false, pageIndex, pageSize);
    }

    /**
     * Returns a page of opens.
     * <p>
     * <a href="http://dev.maileon.com/api/rest-api-1-0/reports/get-opens/?lang=en">Documentation website</a></p>
     *
     * @param fromDate if provided, only the opens after the given date will be returned
     * @param toDate if provided, only the opens before the given date will be returned
     * @param mailingIdFilter multivalued parameter to filter the opens by mailings
     * @param contactIDsFilter multivalued parameter to filter the results by contacts
     * @param contactEmailsFilter multivalued parameter to filter the results by email addresses
     * @param contactExternalIDsFilter multivalued parameter to filter the results by external ids
     * @param formatFilter filters the opens by format. Possible values are: html / text.
     * @param socialNetworkFilter multivalued parameter to filter the opens by the social networks where they occurred
     * @param deviceTypeFilter multivalued parameter to filter the opens by device type. Possible values are: <code>UNKNOWN / COMPUTER / TABLET / MOBILE</code>.
     * @param embedEmailClientInfos ff the set to <code>true</code>, available email client details will be appended to each open
     * @param excludeAnonymousOpens if <code>true</code>, the opens that cannot be mapped to contacts because of permission types or because of unsubscriptions are omitted in the
     * results
     * @param standardFields multivalued parameter to query the value of a standard contact field
     * @param customFields multivalued parameter to query the value of a custom contact field
     * @param embedFieldBackups if <code>true</code>, the corresponding field backups will be returned if available
     * @param pageIndex the index of the result page starting from 1
     * @param pageSize the number of {@link Open} in the resulted page
     * @return page containing the corresponding opens
     * @throws MaileonException
     */
    public Page<Open> getOpens(Long fromDate, Long toDate, List<Long> mailingIdFilter, List<Long> contactIDsFilter, List<String> contactEmailsFilter,
                               List<String> contactExternalIDsFilter, Format formatFilter, List<String> socialNetworkFilter,
                               List<String> deviceTypeFilter, boolean embedEmailClientInfos,
                               boolean excludeAnonymousOpens, List<StandardContactField> standardFields,
                               List<String> customFields, boolean embedFieldBackups, int pageIndex,
                               int pageSize) throws MaileonException {
        validatePaginationParameters(pageIndex, pageSize);
        QueryParameters params = createQueryParameters(pageIndex, pageSize, fromDate, toDate, mailingIdFilter,
                embedFieldBackups);
        addParameters(params, formatFilter, socialNetworkFilter, deviceTypeFilter, embedEmailClientInfos);
        addParameters(params, standardFields, customFields);
        addParameters(params, contactIDsFilter, contactEmailsFilter, contactExternalIDsFilter);
        params.add("exclude_anonymous_opens", excludeAnonymousOpens);
        ResponseWrapper resp = get("reports/opens", params);
        Page<Open> page = new Page<>();
        page.setPageIndex(pageIndex);
        page.setPageSize(pageSize);
        page.setTotalItems(Integer.parseInt(resp.getHeaders().getFirst("X-Items")));
        page.setCountPages(Integer.parseInt(resp.getHeaders().getFirst("X-Pages")));
        Element xml = resp.getEntityAsXml();
        page.setItems(FactsAdaptor.parseOpens(xml));
        return page;
    }

    /**
     * Returns a page of conversions.
     * <p>
     * <a href="http://dev.maileon.com/api/rest-api-1-0/reports/get-conversions/">Documentation website</a></p>
     *
     * @param fromDate if provided, only the opens after the given date will be returned
     * @param toDate if provided, only the opens before the given date will be returned
     * @param mailingIdFilter multivalued parameter to filter the opens by mailings
     * @param contactIDsFilter multivalued parameter to filter the results by contacts
     * @param contactEmailsFilter multivalued parameter to filter the results by email addresses
     * @param contactExternalIDsFilter multivalued parameter to filter the results by external ids
     * @param excludeAnonymousConversions if <code>true</code>, the conversions that cannot be mapped to contacts because of permission types or because of unsubscriptions are omitted in the
     * results
     * @param pageIndex the index of the result page starting from 1
     * @param pageSize the number of {@link Conversion} in the resulted page
     * @return page containing the corresponding conversions
     * @throws MaileonException
     */
    public Page<Conversion> getConversions(Long fromDate, Long toDate, List<Long> mailingIdFilter, List<Long> contactIDsFilter, List<String> contactEmailsFilter,
                               List<String> contactExternalIDsFilter, boolean excludeAnonymousConversions, int pageIndex,
                               int pageSize) throws MaileonException {
        validatePaginationParameters(pageIndex, pageSize);
        QueryParameters params = createQueryParameters(pageIndex, pageSize, fromDate, toDate, mailingIdFilter);
        addParameters(params, contactIDsFilter, contactEmailsFilter, contactExternalIDsFilter);
        params.add("exclude_anonymous_conversions", excludeAnonymousConversions);
        ResponseWrapper resp = get("reports/analytics/conversions", params);
        Page<Conversion> page = new Page<>();
        page.setPageIndex(pageIndex);
        page.setPageSize(pageSize);
        page.setTotalItems(Integer.parseInt(resp.getHeaders().getFirst("X-Items")));
        page.setCountPages(Integer.parseInt(resp.getHeaders().getFirst("X-Pages")));
        Element xml = resp.getEntityAsXml();
        page.setItems(FactsAdaptor.parseConversions(xml));
        return page;
    }

    /**
     * Returns a page of unique conversions.
     * <p>
     * <a href="http://dev.maileon.com/api/rest-api-1-0/reports/get-opens/?lang=en">Documentation website</a></p>
     *
     * @param fromDate if provided, only the opens after the given date will be returned
     * @param toDate if provided, only the opens before the given date will be returned
     * @param mailingIdFilter multivalued parameter to filter the opens by mailings
     * @param contactIDsFilter multivalued parameter to filter the results by contacts
     * @param contactEmailsFilter multivalued parameter to filter the results by email addresses
     * @param contactExternalIDsFilter multivalued parameter to filter the results by external ids
     * @param excludeAnonymousConversions if <code>true</code>, the unique conversions that cannot be mapped to contacts because of permission types or because of unsubscriptions are omitted in the
     * results
     * @param pageIndex the index of the result page starting from 1
     * @param pageSize the number of {@link UniqueConversion} in the resulted page
     * @return page containing the corresponding unique conversions
     * @throws MaileonException
     */
    public Page<UniqueConversion> getUniqueConversions(Long fromDate, Long toDate, List<Long> mailingIdFilter, List<Long> contactIDsFilter, List<String> contactEmailsFilter,
                               List<String> contactExternalIDsFilter, boolean excludeAnonymousConversions, int pageIndex,
                               int pageSize) throws MaileonException {
        validatePaginationParameters(pageIndex, pageSize);
        QueryParameters params = createQueryParameters(pageIndex, pageSize, fromDate, toDate, mailingIdFilter);
        addParameters(params, contactIDsFilter, contactEmailsFilter, contactExternalIDsFilter);
        params.add("exclude_anonymous_conversions", excludeAnonymousConversions);
        ResponseWrapper resp = get("reports/analytics/conversions/unique", params);
        Page<UniqueConversion> page = new Page<>();
        page.setPageIndex(pageIndex);
        page.setPageSize(pageSize);
        page.setTotalItems(Integer.parseInt(resp.getHeaders().getFirst("X-Items")));
        page.setCountPages(Integer.parseInt(resp.getHeaders().getFirst("X-Pages")));
        Element xml = resp.getEntityAsXml();
        page.setItems(FactsAdaptor.parseUniqueConversions(xml));
        return page;
    }

    /**
     * Returns a page of unique opens.
     * <p>
     * <a href="http://dev.maileon.com/api/rest-api-1-0/reports/get-unique-opens/?lang=en">Documentation website</a></p>
     *
     * @param fromDate if provided, only the unique opens after the given date will be returned
     * @param toDate if provided, only the unique opens before the given date will be returned
     * @param mailingIdFilter multivalued parameter to filter the unique opens by mailings
     * @param embedEmailClientInfos ff the set to <code>true</code>, available email client details will be appended to each unique open
     * @param excludeAnonymousOpens if <code>true</code>, the unique opens that cannot be mapped to contacts because of permission types or because of unsubscriptions are omitted
     * in the results
     * @param standardFields multivalued parameter to query the value of a standard contact field
     * @param customFields multivalued parameter to query the value of a custom contact field
     * @param pageIndex the index of the result page starting from 1
     * @param pageSize the number of {@link Open} in the resulted page
     * @return page containing the corresponding unique opens
     * @throws MaileonException
     */
    public Page<Open> getUniqueOpens(Long fromDate, Long toDate, List<Long> mailingIdFilter, boolean embedEmailClientInfos,
            boolean excludeAnonymousOpens, List<StandardContactField> standardFields,
            List<String> customFields, int pageIndex, int pageSize) throws MaileonException {
        return getUniqueOpens(fromDate, toDate, mailingIdFilter, null, null, null, embedEmailClientInfos, excludeAnonymousOpens,
                standardFields, customFields, false, pageIndex, pageSize);
    }

    /**
     * Returns a page of unique opens.
     * <p>
     * <a href="http://dev.maileon.com/api/rest-api-1-0/reports/get-unique-opens/?lang=en">Documentation website</a></p>
     *
     * @param fromDate if provided, only the unique opens after the given date will be returned
     * @param toDate if provided, only the unique opens before the given date will be returned
     * @param mailingIdFilter multivalued parameter to filter the unique opens by mailings
     * @param contactIDsFilter multivalued parameter to filter the results by contacts
     * @param contactEmailsFilter multivalued parameter to filter the results by email addresses
     * @param contactExternalIDsFilter multivalued parameter to filter the results by external ids <code>UNKNOWN / COMPUTER / TABLET / MOBILE</code>.
     * @param embedEmailClientInfos ff the set to <code>true</code>, available email client details will be appended to each unique open
     * @param excludeAnonymousOpens if <code>true</code>, the unique opens that cannot be mapped to contacts because of permission types or because of unsubscriptions are omitted
     * in the results
     * @param standardFields multivalued parameter to query the value of a standard contact field
     * @param customFields multivalued parameter to query the value of a custom contact field
     * @param embedFieldBackups if <code>true</code>, the corresponding field backups will be returned if available
     * @param pageIndex the index of the result page starting from 1
     * @param pageSize the number of {@link Open} in the resulted page
     * @return page containing the corresponding unique opens
     * @throws MaileonException
     */
    public Page<Open> getUniqueOpens(Long fromDate, Long toDate, List<Long> mailingIdFilter, List<Long> contactIDsFilter, List<String> contactEmailsFilter,
            List<String> contactExternalIDsFilter, boolean embedEmailClientInfos,
            boolean excludeAnonymousOpens, List<StandardContactField> standardFields,
            List<String> customFields, boolean embedFieldBackups, int pageIndex,
            int pageSize) throws MaileonException {
        validatePaginationParameters(pageIndex, pageSize);
        QueryParameters params = createQueryParameters(pageIndex, pageSize, fromDate, toDate, mailingIdFilter,
                embedFieldBackups);
        addParameters(params, null, null, null, embedEmailClientInfos);
        addParameters(params, standardFields, customFields);
        addParameters(params, contactIDsFilter, contactEmailsFilter, contactExternalIDsFilter);
        params.add("exclude_anonymous_opens", excludeAnonymousOpens);
        ResponseWrapper resp = get("reports/opens/unique", params);
        Page<Open> page = new Page<>();
        page.setPageIndex(pageIndex);
        page.setPageSize(pageSize);
        page.setTotalItems(Integer.parseInt(resp.getHeaders().getFirst("X-Items")));
        page.setCountPages(Integer.parseInt(resp.getHeaders().getFirst("X-Pages")));
        Element xml = resp.getEntityAsXml();
        page.setItems(FactsAdaptor.parseOpens(xml));
        return page;
    }

    /**
     * Returns the number of opens.
     *
     * @param fromDate if provided, only the opens after the given date will be returned
     * @param toDate if provided, only the opens before the given date will be returned
     * @param mailingIdFilter multivalued parameter to filter the opens by mailings
     * @param excludeAnonymousOpens if <code>true</code>, the opens that cannot be mapped to contacts because of permission types or because of unsubscriptions are omitted in the
     * results
     * @return the count of opens
     * @throws MaileonException
     */
    public int getOpensCount(Long fromDate, Long toDate, List<Long> mailingIdFilter, boolean excludeAnonymousOpens) throws MaileonException {
        return getOpensCount(fromDate, toDate, mailingIdFilter, null, null, null, null, null, null, excludeAnonymousOpens);
    }

    /**
     * Returns the number of opens.
     *
     * @param fromDate if provided, only the opens after the given date will be returned
     * @param toDate if provided, only the opens before the given date will be returned
     * @param mailingIdFilter multivalued parameter to filter the opens by mailings
     * @param contactIDsFilter multivalued parameter to filter the results by contacts
     * @param contactEmailsFilter multivalued parameter to filter the results by email addresses
     * @param contactExternalIDsFilter multivalued parameter to filter the results by external ids
     * @param formatFilter filters the opens by format. Possible values are: html / text.
     * @param socialNetworkFilter multivalued parameter to filter the opens by the social networks where they occurred
     * @param deviceTypeFilter multivalued parameter to filter the opens by device type. Possible values are: <code>UNKNOWN / COMPUTER / TABLET / MOBILE</code>.
     * @param excludeAnonymousOpens if <code>true</code>, the opens that cannot be mapped to contacts because of permission types or because of unsubscriptions are omitted in the
     * results
     * @return the count of opens
     * @throws MaileonException
     */
    public int getOpensCount(Long fromDate, Long toDate, List<Long> mailingIdFilter, List<Long> contactIDsFilter, List<String> contactEmailsFilter,
            List<String> contactExternalIDsFilter, Format formatFilter, List<String> socialNetworkFilter,
            List<String> deviceTypeFilter, boolean excludeAnonymousOpens) throws MaileonException {
        QueryParameters params = new QueryParameters();
        params.add("from_date", fromDate);
        params.add("to_date", toDate);
        if (mailingIdFilter != null) {
            for (Long id : mailingIdFilter) {
                params.add("mailing_id", id);
            }
        }
        addParameters(params, formatFilter, socialNetworkFilter, deviceTypeFilter, null);
        addParameters(params, contactIDsFilter, contactEmailsFilter, contactExternalIDsFilter);
        params.add("exclude_anonymous_opens", excludeAnonymousOpens);

        Element resp = get("reports/opens/count", params).getEntityAsXml();
        try {
            return Integer.parseInt(resp.getText());
        } catch (Throwable t) {
            throw new MaileonClientException("unexpected response format", t);
        }
    }

    /**
     * Returns the number of unique opens.
     * <p>
     * <a href="http://dev.maileon.com/api/rest-api-1-0/reports/get-unique-opens/?lang=en">Documentation website</a></p>
     *
     * @param fromDate if provided, only the unique opens after the given date will be returned
     * @param toDate if provided, only the unique opens before the given date will be returned
     * @param mailingIdFilter multivalued parameter to filter the unique opens by mailings
     * @param excludeAnonymousOpens if <code>true</code>, the unique opens that cannot be mapped to contacts because of permission types or because of unsubscriptions are omitted
     * in the results
     * @return the count of unique opens
     * @throws MaileonException
     */
    public int getUniqueOpensCount(Long fromDate, Long toDate, List<Long> mailingIdFilter, boolean excludeAnonymousOpens) throws MaileonException {
        return getUniqueOpensCount(fromDate, toDate, mailingIdFilter, null, null, null, null, null, null, excludeAnonymousOpens);
    }

    /**
     * Returns the number of unique opens.
     * <p>
     * <a href="http://dev.maileon.com/api/rest-api-1-0/reports/get-unique-opens/?lang=en">Documentation website</a></p>
     *
     * @param fromDate if provided, only the unique opens after the given date will be returned
     * @param toDate if provided, only the unique opens before the given date will be returned
     * @param mailingIdFilter multivalued parameter to filter the unique opens by mailings
     * @param contactIDsFilter multivalued parameter to filter the results by contacts
     * @param contactEmailsFilter multivalued parameter to filter the results by email addresses
     * @param contactExternalIDsFilter multivalued parameter to filter the results by external ids
     * @param formatFilter filters the unique opens by format. Possible values are: html / text.
     * @param socialNetworkFilter multivalued parameter to filter the unique opens by the social networks where they occurred
     * @param deviceTypeFilter multivalued parameter to filter the unique opens by device type. Possible values are: <code>UNKNOWN / COMPUTER / TABLET / MOBILE</code>.
     * @param excludeAnonymousOpens if <code>true</code>, the unique opens that cannot be mapped to contacts because of permission types or because of unsubscriptions are omitted
     * in the results
     * @return the count of unique opens
     * @throws MaileonException
     */
    public int getUniqueOpensCount(Long fromDate, Long toDate, List<Long> mailingIdFilter, List<Long> contactIDsFilter, List<String> contactEmailsFilter,
            List<String> contactExternalIDsFilter, Format formatFilter, List<String> socialNetworkFilter,
            List<String> deviceTypeFilter, boolean excludeAnonymousOpens) throws MaileonException {
        QueryParameters params = new QueryParameters();
        params.add("from_date", fromDate);
        params.add("to_date", toDate);
        if (mailingIdFilter != null) {
            for (Long id : mailingIdFilter) {
                params.add("mailing_id", id);
            }
        }
        addParameters(params, formatFilter, socialNetworkFilter, deviceTypeFilter, null);
        addParameters(params, contactIDsFilter, contactEmailsFilter, contactExternalIDsFilter);
        params.add("exclude_anonymous_opens", excludeAnonymousOpens);

        Element resp = get("reports/opens/unique/count", params).getEntityAsXml();
        try {
            return Integer.parseInt(resp.getText());
        } catch (Throwable t) {
            throw new MaileonClientException("unexpected response format", t);
        }
    }

    /**
     * Returns a page of contacts that received at least one mailing.
     * <p>
     * <a href="http://dev.maileon.com/api/rest-api-1-0/reports/get-recipients/?lang=en">Documentation website</a></p>
     *
     * @param fromDate if provided, only the opens after the given date will be returned
     * @param toDate if provided, only the opens before the given date will be returned
     * @param mailingIdFilter multivalued parameter to filter the opens by mailings
     * @param excludeDeletedRecipients if <code>true</code>, the recipients that have been removed from maileon will be excluded
     * @param standardFields multivalued parameter to query the value of a standard contact field
     * @param customFields multivalued parameter to query the value of a custom contact field
     * @param pageIndex the index of the result page starting from 1
     * @param pageSize the number of {@link Open} in the resulted page
     * @return page containing the corresponding opens
     * @throws MaileonException
     */
    public Page<Recipient> getRecipients(Long fromDate, Long toDate, List<Long> mailingIdFilter, boolean excludeDeletedRecipients, List<StandardContactField> standardFields,
            List<String> customFields, int pageIndex, int pageSize) throws MaileonException {
        return getRecipients(fromDate, toDate, mailingIdFilter, null, null, null, excludeDeletedRecipients, standardFields, customFields, false, pageIndex, pageSize);
    }

    /**
     * Returns a page of contacts that received at least one mailing.
     * <p>
     * <a href="http://dev.maileon.com/api/rest-api-1-0/reports/get-recipients/?lang=en">Documentation website</a></p>
     *
     * @param fromDate if provided, only the opens after the given date will be returned
     * @param toDate if provided, only the opens before the given date will be returned
     * @param mailingIdFilter multivalued parameter to filter the opens by mailings
     * @param contactIDsFilter multivalued parameter to filter the results by contacts
     * @param contactEmailsFilter multivalued parameter to filter the results by email addresses
     * @param contactExternalIDsFilter multivalued parameter to filter the results by external ids
     * @param excludeDeletedRecipients if <code>true</code>, the recipients that have been removed from maileon will be excluded
     * @param standardFields multivalued parameter to query the value of a standard contact field
     * @param customFields multivalued parameter to query the value of a custom contact field
     * @param embedFieldBackups if <code>true</code>, the corresponding field backups will be returned if available
     * @param pageIndex the index of the result page starting from 1
     * @param pageSize the number of {@link Open} in the resulted page
     * @return page containing the corresponding opens
     * @throws MaileonException
     */
    public Page<Recipient> getRecipients(Long fromDate, Long toDate, List<Long> mailingIdFilter, List<Long> contactIDsFilter, List<String> contactEmailsFilter,
            List<String> contactExternalIDsFilter, boolean excludeDeletedRecipients, List<StandardContactField> standardFields,
            List<String> customFields, boolean embedFieldBackups, int pageIndex,
            int pageSize) throws MaileonException {
        validatePaginationParameters(pageIndex, pageSize);
        QueryParameters params = createQueryParameters(pageIndex, pageSize, fromDate, toDate, mailingIdFilter,
                embedFieldBackups);
        addParameters(params, null, null, null, null);
        addParameters(params, standardFields, customFields);
        addParameters(params, contactIDsFilter, contactEmailsFilter, contactExternalIDsFilter);
        params.add("exclude_deleted_recipients", excludeDeletedRecipients);
        ResponseWrapper resp = get("reports/recipients", params);
        Page<Recipient> page = new Page<>();
        page.setPageIndex(pageIndex);
        page.setPageSize(pageSize);
        page.setTotalItems(Integer.parseInt(resp.getHeaders().getFirst("X-Items")));
        page.setCountPages(Integer.parseInt(resp.getHeaders().getFirst("X-Pages")));
        Element xml = resp.getEntityAsXml();
        page.setItems(FactsAdaptor.parseRecipients(xml));
        return page;
    }

        /**
     * Returns the number of recipients.
     *
     * @param fromDate if provided, only the opens after the given date will be returned
     * @param toDate if provided, only the opens before the given date will be returned
     * @param mailingIdFilter multivalued parameter to filter the opens by mailings
     * @param excludeDeletedRecipients if <code>true</code>, the recipients that have been removed from maileon will be excluded
     * @return the count of the recipients
     * @throws MaileonException
     */
    public int getRecipientsCount(Long fromDate, Long toDate, List<Long> mailingIdFilter, boolean excludeDeletedRecipients) throws MaileonException {
        return getRecipientsCount(fromDate, toDate, mailingIdFilter, null, null, null, excludeDeletedRecipients);
    }

    /**
     * Returns the number of recipients.
     *
     * @param fromDate if provided, only the opens after the given date will be returned
     * @param toDate if provided, only the opens before the given date will be returned
     * @param mailingIdFilter multivalued parameter to filter the opens by mailings
     * @param contactIDsFilter multivalued parameter to filter the results by contacts
     * @param contactEmailsFilter multivalued parameter to filter the results by email addresses
     * @param contactExternalIDsFilter multivalued parameter to filter the results by external ids
     * @param excludeDeletedRecipients if <code>true</code>, the recipients that have been removed from maileon will be excluded
     * @return the count of the recipients
     * @throws MaileonException
     */
    public int getRecipientsCount(Long fromDate, Long toDate, List<Long> mailingIdFilter, List<Long> contactIDsFilter, List<String> contactEmailsFilter,
            List<String> contactExternalIDsFilter, boolean excludeDeletedRecipients) throws MaileonException {
        QueryParameters params = new QueryParameters();
        params.add("from_date", fromDate);
        params.add("to_date", toDate);
        params.add("exclude_deleted_recipients", excludeDeletedRecipients);
        if (mailingIdFilter != null) {
            for (Long id : mailingIdFilter) {
                params.add("mailing_id", id);
            }
        }
        addParameters(params, contactIDsFilter, contactEmailsFilter, contactExternalIDsFilter);

        Element resp = get("reports/recipients/count", params).getEntityAsXml();
        try {
            return Integer.parseInt(resp.getText());
        } catch (Throwable t) {
            throw new MaileonClientException("unexpected response format", t);
        }
    }

    /**
     * Returns a page of clicks.
     * <p>
     * <a href="http://dev.maileon.com/api/rest-api-1-0/reports/draft-get-clicks/?lang=en">Documentation website</a></p>
     *
     * @param fromDate if provided, only the clicks after the given date will be returned
     * @param toDate if provided, only the clicks before the given date will be returned
     * @param mailingIdFilter multivalued parameter to filter the clicks by mailings
     * @param embedEmailClientInfos ff the set to <code>true</code>, available email client details will be appended to each click
     * @param excludeAnonymousClicks if <code>true</code>, the clicks that cannot be mapped to contacts because of permission types or because of unsubscriptions are omitted in the
     * results
     * @param standardFields multivalued parameter to query the value of a standard contact field
     * @param customFields multivalued parameter to query the value of a custom contact field
     * @param pageIndex the index of the result page starting from 1
     * @param pageSize the number of {@link Click} in the resulted page
     * @return page containing the corresponding clicks
     * @throws MaileonException
     */
    public Page<Click> getClicks(Long fromDate, Long toDate, List<Long> mailingIdFilter, boolean embedEmailClientInfos,
            boolean excludeAnonymousClicks, List<StandardContactField> standardFields,
            List<String> customFields, int pageIndex, int pageSize) throws MaileonException {
        return getClicks(fromDate, toDate, mailingIdFilter, null, null, null, null, null, null, null, null, null,
                embedEmailClientInfos, excludeAnonymousClicks, standardFields,
                customFields, false, pageIndex, pageSize);
    }

    /**
     * Returns a page of clicks.
     * <p>
     * <a href="http://dev.maileon.com/api/rest-api-1-0/reports/draft-get-clicks/?lang=en">Documentation website</a></p>
     *
     * @param fromDate if provided, only the clicks after the given date will be returned
     * @param toDate if provided, only the clicks before the given date will be returned
     * @param mailingIdFilter multivalued parameter to filter the clicks by mailings
     * @param contactIDsFilter multivalued parameter to filter the results by contacts
     * @param contactEmailsFilter multivalued parameter to filter the results by email addresses
     * @param contactExternalIDsFilter multivalued parameter to filter the results by external ids
     * @param formatFilter filters the clicks by format. Possible values are: html / text.
     * @param linkIdFilter multivalued parameter to filter the clicks by links
     * @param linkUrlFilter filters the clicks by link url
     * @param linkTagFilter multivalued parameter to filter the clicks by tags associated to the links
     * @param socialNetworkFilter multivalued parameter to filter the clicks by the social networks where they occurred
     * @param deviceTypeFilter multivalued parameter to filter the clicks by device type. Possible values are: <code>UNKNOWN / COMPUTER / TABLET / MOBILE</code>.
     * @param embedEmailClientInfos ff the set to <code>true</code>, available email client details will be appended to each click
     * @param excludeAnonymousClicks if <code>true</code>, the clicks that cannot be mapped to contacts because of permission types or because of unsubscriptions are omitted in the
     * results
     * @param standardFields multivalued parameter to query the value of a standard contact field
     * @param customFields multivalued parameter to query the value of a custom contact field
     * @param embedFieldBackups if <code>true</code>, the corresponding field backups will be returned if available
     * @param pageIndex the index of the result page starting from 1
     * @param pageSize the number of {@link Click} in the resulted page
     * @return page containing the corresponding clicks
     * @throws MaileonException
     */
    public Page<Click> getClicks(Long fromDate, Long toDate, List<Long> mailingIdFilter, List<Long> contactIDsFilter, List<String> contactEmailsFilter,
            List<String> contactExternalIDsFilter, Format formatFilter, List<Long> linkIdFilter, String linkUrlFilter,
            List<String> linkTagFilter, List<String> socialNetworkFilter,
            List<String> deviceTypeFilter, boolean embedEmailClientInfos,
            boolean excludeAnonymousClicks, List<StandardContactField> standardFields,
            List<String> customFields, boolean embedFieldBackups, int pageIndex,
            int pageSize) throws MaileonException {
        validatePaginationParameters(pageIndex, pageSize);
        QueryParameters params = createQueryParameters(pageIndex, pageSize, fromDate, toDate, mailingIdFilter,
                embedFieldBackups);
        addParameters(params, formatFilter, socialNetworkFilter, deviceTypeFilter, embedEmailClientInfos);
        addParameters(params, standardFields, customFields);
        addParameters(params, contactIDsFilter, contactEmailsFilter, contactExternalIDsFilter);
        if (linkIdFilter != null) {
            for (Long linkId : linkIdFilter) {
                params.add("link_id", linkId);
            }
        }
        params.add("link_url", linkUrlFilter);
        if (linkTagFilter != null) {
            for (String linkTag : linkTagFilter) {
                params.add("link_tag", linkTag);
            }
        }
        params.add("exclude_anonymous_clicks", excludeAnonymousClicks);
        ResponseWrapper resp = get("reports/clicks", params);
        Page<Click> page = new Page<>();
        page.setPageIndex(pageIndex);
        page.setPageSize(pageSize);
        page.setTotalItems(Integer.parseInt(resp.getHeaders().getFirst("X-Items")));
        page.setCountPages(Integer.parseInt(resp.getHeaders().getFirst("X-Pages")));
        Element xml = resp.getEntityAsXml();
        page.setItems(FactsAdaptor.parseClicks(xml));
        return page;
    }

    /**
     * Returns a page of unique clicks.
     * <p>
     * <a href="http://dev.maileon.com/api/rest-api-1-0/reports/get-unique-clicks/?lang=en">Documentation website</a></p>
     *
     * @param fromDate if provided, only the unique clicks after the given date will be returned
     * @param toDate if provided, only the unique clicks before the given date will be returned
     * @param mailingIdFilter multivalued parameter to filter the unique clicks by mailings
     * @param embedEmailClientInfos ff the set to <code>true</code>, available email client details will be appended to each unique click
     * @param excludeAnonymousClicks if <code>true</code>, the unique clicks that cannot be mapped to contacts because of permission types or because of unsubscriptions are omitted
     * in the results
     * @param standardFields multivalued parameter to query the value of a standard contact field
     * @param customFields multivalued parameter to query the value of a custom contact field
     * @param pageIndex the index of the result page starting from 1
     * @param pageSize the number of {@link Click} in the resulted page
     * @return page containing the corresponding unique clicks
     * @throws MaileonException
     */
    public Page<Click> getUniqueClicks(Long fromDate, Long toDate, List<Long> mailingIdFilter, boolean embedEmailClientInfos,
            boolean excludeAnonymousClicks, List<StandardContactField> standardFields,
            List<String> customFields, int pageIndex, int pageSize) throws MaileonException {
        return getUniqueClicks(fromDate, toDate, mailingIdFilter, null, null, null, embedEmailClientInfos, excludeAnonymousClicks,
                standardFields, customFields, false, pageIndex, pageSize);
    }

    /**
     * Returns a page of unique clicks.
     * <p>
     * <a href="http://dev.maileon.com/api/rest-api-1-0/reports/get-unique-clicks/?lang=en">Documentation website</a></p>
     *
     * @param fromDate if provided, only the unique clicks after the given date will be returned
     * @param toDate if provided, only the unique clicks before the given date will be returned
     * @param mailingIdFilter multivalued parameter to filter the unique clicks by mailings
     * @param contactIDsFilter multivalued parameter to filter the results by contacts
     * @param contactEmailsFilter multivalued parameter to filter the results by email addresses
     * @param contactExternalIDsFilter multivalued parameter to filter the results by external ids
     * @param embedEmailClientInfos ff the set to <code>true</code>, available email client details will be appended to each unique click
     * @param excludeAnonymousClicks if <code>true</code>, the unique clicks that cannot be mapped to contacts because of permission types or because of unsubscriptions are omitted
     * in the results
     * @param standardFields multivalued parameter to query the value of a standard contact field
     * @param customFields multivalued parameter to query the value of a custom contact field
     * @param embedFieldBackups if <code>true</code>, the corresponding field backups will be returned if available
     * @param pageIndex the index of the result page starting from 1
     * @param pageSize the number of {@link Click} in the resulted page
     * @return page containing the corresponding unique clicks
     * @throws MaileonException
     */
    public Page<Click> getUniqueClicks(Long fromDate, Long toDate, List<Long> mailingIdFilter, List<Long> contactIDsFilter, List<String> contactEmailsFilter,
            List<String> contactExternalIDsFilter, boolean embedEmailClientInfos,
            boolean excludeAnonymousClicks, List<StandardContactField> standardFields,
            List<String> customFields, boolean embedFieldBackups, int pageIndex,
            int pageSize) throws MaileonException {
        validatePaginationParameters(pageIndex, pageSize);
        QueryParameters params = createQueryParameters(pageIndex, pageSize, fromDate, toDate, mailingIdFilter,
                embedFieldBackups);
        addParameters(params, null, null, null, embedEmailClientInfos);
        addParameters(params, standardFields, customFields);
        addParameters(params, contactIDsFilter, contactEmailsFilter, contactExternalIDsFilter);

        params.add("exclude_anonymous_clicks", excludeAnonymousClicks);
        ResponseWrapper resp = get("reports/clicks/unique", params);
        Page<Click> page = new Page<>();
        page.setPageIndex(pageIndex);
        page.setPageSize(pageSize);
        page.setTotalItems(Integer.parseInt(resp.getHeaders().getFirst("X-Items")));
        page.setCountPages(Integer.parseInt(resp.getHeaders().getFirst("X-Pages")));
        Element xml = resp.getEntityAsXml();
        page.setItems(FactsAdaptor.parseClicks(xml));
        return page;
    }

    /**
     * Returns the number of clicks.
     *
     * @param fromDate if provided, only the clicks after the given date will be returned
     * @param toDate if provided, only the clicks before the given date will be returned
     * @param mailingIdFilter multivalued parameter to filter the clicks by mailings
     * @param excludeAnonymousClicks if <code>true</code>, the clicks that cannot be mapped to contacts because of permission types or because of unsubscriptions are omitted in the
     * results
     * @return the count of clicks
     * @throws MaileonException
     */
    public int getClicksCount(Long fromDate, Long toDate, List<Long> mailingIdFilter,
            boolean excludeAnonymousClicks) throws MaileonException {
        return getClicksCount(fromDate, toDate, mailingIdFilter, null, null, null, null, null, null, null, null, null, excludeAnonymousClicks);
    }

    /**
     * Returns the number of clicks.
     *
     * @param fromDate if provided, only the clicks after the given date will be returned
     * @param toDate if provided, only the clicks before the given date will be returned
     * @param mailingIdFilter multivalued parameter to filter the clicks by mailings
     * @param contactIDsFilter multivalued parameter to filter the results by contacts
     * @param contactEmailsFilter multivalued parameter to filter the results by email addresses
     * @param contactExternalIDsFilter multivalued parameter to filter the results by external ids
     * @param formatFilter filters the clicks by format. Possible values are: html / text.
     * @param linkIdFilter multivalued parameter to filter the clicks by links
     * @param linkUrlFilter filters the clicks by link url
     * @param linkTagFilter multivalued parameter to filter the clicks by tags associated to the links
     * @param socialNetworkFilter multivalued parameter to filter the clicks by the social networks where they occurred
     * @param deviceTypeFilter multivalued parameter to filter the clicks by device type. Possible values are: <code>UNKNOWN / COMPUTER / TABLET / MOBILE</code>.
     * @param excludeAnonymousClicks if <code>true</code>, the clicks that cannot be mapped to contacts because of permission types or because of unsubscriptions are omitted in the
     * results
     * @return the count of clicks
     * @throws MaileonException
     */
    public int getClicksCount(Long fromDate, Long toDate, List<Long> mailingIdFilter, List<Long> contactIDsFilter, List<String> contactEmailsFilter,
            List<String> contactExternalIDsFilter, Format formatFilter, List<Long> linkIdFilter, String linkUrlFilter,
            List<String> linkTagFilter, List<String> socialNetworkFilter,
            List<String> deviceTypeFilter,
            boolean excludeAnonymousClicks) throws MaileonException {
        QueryParameters params = new QueryParameters();
        params.add("from_date", fromDate);
        params.add("to_date", toDate);
        if (mailingIdFilter != null) {
            for (Long id : mailingIdFilter) {
                params.add("mailing_id", id);
            }
        }
        addParameters(params, formatFilter, socialNetworkFilter, deviceTypeFilter, null);
        addParameters(params, contactIDsFilter, contactEmailsFilter, contactExternalIDsFilter);
        if (linkIdFilter != null) {
            for (Long linkId : linkIdFilter) {
                params.add("link_id", linkId);
            }
        }
        params.add("link_url", linkUrlFilter);
        if (linkTagFilter != null) {
            for (String linkTag : linkTagFilter) {
                params.add("link_tag", linkTag);
            }
        }
        params.add("exclude_anonymous_clicks", excludeAnonymousClicks);

        Element resp = get("reports/clicks/count", params).getEntityAsXml();
        try {
            return Integer.parseInt(resp.getText());
        } catch (Throwable t) {
            throw new MaileonClientException("unexpected response format", t);
        }
    }

    /**
     * Returns the number of unique clicks.
     *
     * @param fromDate if provided, only the unique clicks after the given date will be returned
     * @param toDate if provided, only the unique clicks before the given date will be returned
     * @param mailingIdFilter multivalued parameter to filter the unique clicks by mailings
     * @param excludeAnonymousClicks if <code>true</code>, the unique clicks that cannot be mapped to contacts because of permission types or because of unsubscriptions are omitted
     * in the results
     * @return the count of clicks
     * @throws MaileonException
     */
    public int getUniqueClicksCount(Long fromDate, Long toDate, List<Long> mailingIdFilter, boolean excludeAnonymousClicks) throws MaileonException {
        return getUniqueClicksCount(fromDate, toDate, mailingIdFilter, null, null, null, excludeAnonymousClicks);
    }

    /**
     * Returns the number of unique clicks.
     *
     * @param fromDate if provided, only the unique clicks after the given date will be returned
     * @param toDate if provided, only the unique clicks before the given date will be returned
     * @param mailingIdFilter multivalued parameter to filter the unique clicks by mailings
     * @param contactIDsFilter multivalued parameter to filter the results by contacts
     * @param contactEmailsFilter multivalued parameter to filter the results by email addresses
     * @param contactExternalIDsFilter multivalued parameter to filter the results by external ids
     * @param excludeAnonymousClicks if <code>true</code>, the unique clicks that cannot be mapped to contacts because of permission types or because of unsubscriptions are omitted
     * in the results
     * @return the count of clicks
     * @throws MaileonException
     */
    public int getUniqueClicksCount(Long fromDate, Long toDate, List<Long> mailingIdFilter, List<Long> contactIDsFilter, List<String> contactEmailsFilter,
            List<String> contactExternalIDsFilter, boolean excludeAnonymousClicks) throws MaileonException {
        QueryParameters params = new QueryParameters();
        params.add("from_date", fromDate);
        params.add("to_date", toDate);
        if (mailingIdFilter != null) {
            for (Long id : mailingIdFilter) {
                params.add("mailing_id", id);
            }
        }
        addParameters(params, contactIDsFilter, contactEmailsFilter, contactExternalIDsFilter);
        params.add("exclude_anonymous_clicks", excludeAnonymousClicks);

        Element resp = get("reports/clicks/unique/count", params).getEntityAsXml();
        try {
            return Integer.parseInt(resp.getText());
        } catch (Throwable t) {
            throw new MaileonClientException("unexpected response format", t);
        }
    }

    /**
     * Returns a page of bounces.
     * <p>
     * <a href="http://dev.maileon.com/api/rest-api-1-0/reports/get-bounces/?lang=en">Documentation website</a></p>
     *
     * @param fromDate if provided, only the bounces after the given date will be returned
     * @param toDate if provided, only the bounces before the given date will be returned
     * @param mailingIdFilter multivalued parameter to filter the bounces by mailings
     * @param excludeAnonymousBounces if <code>true</code>, the bounces that cannot be mapped to contacts because of permission types or because unsubscriptions are omitted in the
     * results
     * @param standardFields multivalued parameter to query the value of a standard contact field
     * @param customFields multivalued parameter to query the value of a custom contact field
     * @param pageIndex the index of the result page starting from 1
     * @param pageSize the number of {@link Bounce} in the resulted page
     * @return page containing the corresponding bounces
     * @throws MaileonException
     */
    public Page<Bounce> getBounces(Long fromDate, Long toDate, List<Long> mailingIdFilter, boolean excludeAnonymousBounces, List<StandardContactField> standardFields,
            List<String> customFields, int pageIndex, int pageSize) throws MaileonException {
        return getBounces(fromDate, toDate, mailingIdFilter, null, null, null, null, null, null, excludeAnonymousBounces, standardFields, customFields, false, pageIndex, pageSize);
    }

    /**
     * Returns a page of bounces.
     * <p>
     * <a href="http://dev.maileon.com/api/rest-api-1-0/reports/get-bounces/?lang=en">Documentation website</a></p>
     *
     * @param fromDate if provided, only the bounces after the given date will be returned
     * @param toDate if provided, only the bounces before the given date will be returned
     * @param mailingIdFilter multivalued parameter to filter the bounces by mailings
     * @param contactIDsFilter multivalued parameter to filter the results by contacts
     * @param contactEmailsFilter multivalued parameter to filter the results by email addresses
     * @param contactExternalIDsFilter multivalued parameter to filter the results by external ids
     * @param statusCodeFilter filters the bounces by status codes. Status codes follow the pattern digit.digit.digit (e.g. 5.0.0)
     * @param typeFilter filters the bounces by type: permanent / transient
     * @param sourceFilter filters the bounces by their source: mta-listener / reply
     * @param excludeAnonymousBounces if <code>true</code>, the bounces that cannot be mapped to contacts because of permission types or because unsubscriptions are omitted in the
     * results
     * @param standardFields multivalued parameter to query the value of a standard contact field
     * @param customFields multivalued parameter to query the value of a custom contact field
     * @param embedFieldBackups if <code>true</code>, the corresponding field backups will be returned if available
     * @param pageIndex the index of the result page starting from 1
     * @param pageSize the number of {@link Bounce} in the resulted page
     * @return page containing the corresponding bounces
     * @throws MaileonException
     */
    public Page<Bounce> getBounces(Long fromDate, Long toDate, List<Long> mailingIdFilter, List<Long> contactIDsFilter, List<String> contactEmailsFilter,
            List<String> contactExternalIDsFilter, List<String> statusCodeFilter, String typeFilter, String sourceFilter,
            boolean excludeAnonymousBounces, List<StandardContactField> standardFields,
            List<String> customFields, boolean embedFieldBackups, int pageIndex,
            int pageSize) throws MaileonException {
        validatePaginationParameters(pageIndex, pageSize);
        QueryParameters params = createQueryParameters(pageIndex, pageSize, fromDate, toDate, mailingIdFilter,
                embedFieldBackups);
        addParameters(params, standardFields, customFields);
        addParameters(params, contactIDsFilter, contactEmailsFilter, contactExternalIDsFilter);
        if (statusCodeFilter != null) {
            for (String sc : statusCodeFilter) {
                params.add("status_code", sc);
            }
        }
        params.add("type", typeFilter);
        params.add("source", sourceFilter);
        params.add("exclude_anonymous_bounces", excludeAnonymousBounces);
        ResponseWrapper resp = get("reports/bounces", params);
        Page<Bounce> page = new Page<>();
        page.setPageIndex(pageIndex);
        page.setPageSize(pageSize);
        page.setTotalItems(Integer.parseInt(resp.getHeaders().getFirst("X-Items")));
        page.setCountPages(Integer.parseInt(resp.getHeaders().getFirst("X-Pages")));
        Element xml = resp.getEntityAsXml();
        page.setItems(FactsAdaptor.parseBounces(xml));
        return page;
    }

    /**
     * Returns a page of unsubscribers.
     * <p>
     * <a href="http://dev.maileon.com/api/rest-api-1-0/reports/get-unsubscriptions/?lang=en">Documentation website</a></p>
     *
     * @param fromDate if provided, only the unsubscribers after the given date will be returned
     * @param toDate if provided, only the unsubscribers before the given date will be returned
     * @param mailingIdFilter multivalued parameter to filter the unsubscribers by mailings
     * @param pageIndex the index of the result page starting from 1
     * @param pageSize the number of {@link Unsubscriber} in the resulted page
     * @return page containing the corresponding unsubsribers
     * @throws MaileonException
     */
    public Page<Unsubscriber> getUnsubscribers(Long fromDate, Long toDate, List<Long> mailingIdFilter, int pageIndex, int pageSize) throws MaileonException {
        return getUnsubscribers(fromDate, toDate, mailingIdFilter, null, null, null, null, false, pageIndex, pageSize);
    }

    /**
     * Returns a page of unsubscribers.
     * <p>
     * <a href="http://dev.maileon.com/api/rest-api-1-0/reports/get-unsubscriptions/?lang=en">Documentation website</a></p>
     *
     * @param fromDate if provided, only the unsubscribers after the given date will be returned
     * @param toDate if provided, only the unsubscribers before the given date will be returned
     * @param mailingIdFilter multivalued parameter to filter the unsubscribers by mailings
     * @param contactIDsFilter multivalued parameter to filter the results by contacts
     * @param contactEmailsFilter multivalued parameter to filter the results by email addresses
     * @param contactExternalIDsFilter multivalued parameter to filter the results by external ids
     * @param sourceFilter filters the unsubscriptions by their source. The source can be an unsubscription link (link), a reply mail (reply) or other.
     * @param embedFieldBackups if <code>true</code>, the corresponding field backups will be returned if available
     * @param pageIndex the index of the result page starting from 1
     * @param pageSize the number of {@link Unsubscriber} in the resulted page
     * @return page containing the corresponding unsubsribers
     * @throws MaileonException
     */
    public Page<Unsubscriber> getUnsubscribers(Long fromDate, Long toDate, List<Long> mailingIdFilter, List<Long> contactIDsFilter, List<String> contactEmailsFilter,
            List<String> contactExternalIDsFilter, String sourceFilter, boolean embedFieldBackups, int pageIndex, int pageSize) throws MaileonException {
        validatePaginationParameters(pageIndex, pageSize);
        QueryParameters params = createQueryParameters(pageIndex, pageSize, fromDate, toDate, mailingIdFilter,
                embedFieldBackups);
        if (sourceFilter != null && sourceFilter.length() > 0) {
            params.add("source", sourceFilter);
        }
        addParameters(params, contactIDsFilter, contactEmailsFilter, contactExternalIDsFilter);
        ResponseWrapper resp = get("reports/unsubscriptions", params);
        Page<Unsubscriber> page = new Page<>();
        page.setPageIndex(pageIndex);
        page.setPageSize(pageSize);
        page.setTotalItems(Integer.parseInt(resp.getHeaders().getFirst("X-Items")));
        page.setCountPages(Integer.parseInt(resp.getHeaders().getFirst("X-Pages")));
        Element xml = resp.getEntityAsXml();
        page.setItems(FactsAdaptor.parseUnsubscriptions(xml));
        return page;
    }

    /**
     * Returns a page of contacts that subscribed for the given account.
     * <p>
     * <a href="http://dev.maileon.com/api/rest-api-1-0/reports/get-subscribers/?lang=en">Documentation website</a></p>
     *
     * @param fromDate if provided, only the subscribers after the given date will be returned
     * @param toDate if provided, only the subscribers before the given date will be returned
     * @param mailingIdFilter multivalued parameter to filter the subscribers by mailings
     * @param contactIDsFilter multivalued parameter to filter the results by contacts
     * @param contactEmailsFilter multivalued parameter to filter the results by email addresses
     * @param contactExternalIDsFilter multivalued parameter to filter the results by external ids
     * @param excludeAnonymousClicks if <code>true</code>, the subscribers that cannot be mapped to contacts because of permission types or because of unsubscriptions are omitted
     * in the results
     * @param standardFields multivalued parameter to query the value of a standard contact field
     * @param customFields multivalued parameter to query the value of a custom contact field
     * @param embedFieldBackups if <code>true</code>, the corresponding field backups will be returned if available
     * @param pageIndex the index of the result page starting from 1
     * @param pageSize the number of {@link Subscriber} in the resulted page
     * @return page containing the corresponding subsribers
     * @throws MaileonException
     */
    public Page<Subscriber> getSubscribers(Long fromDate, Long toDate, List<Long> mailingIdFilter, List<Long> contactIDsFilter, List<String> contactEmailsFilter,
            List<String> contactExternalIDsFilter, boolean excludeAnonymousClicks, List<StandardContactField> standardFields,
            List<String> customFields, boolean embedFieldBackups, int pageIndex,
            int pageSize) throws MaileonException {
        validatePaginationParameters(pageIndex, pageSize);
        QueryParameters params = createQueryParameters(pageIndex, pageSize, fromDate, toDate, mailingIdFilter,
                embedFieldBackups);
        addParameters(params, standardFields, customFields);
        addParameters(params, contactIDsFilter, contactEmailsFilter, contactExternalIDsFilter);
        params.add("exclude_anonymous_clicks", excludeAnonymousClicks);
        ResponseWrapper resp = get("reports/subscribers", params);
        Page<Subscriber> page = new Page<>();
        page.setPageIndex(pageIndex);
        page.setPageSize(pageSize);
        page.setTotalItems(Integer.parseInt(resp.getHeaders().getFirst("X-Items")));
        page.setCountPages(Integer.parseInt(resp.getHeaders().getFirst("X-Pages")));
        Element xml = resp.getEntityAsXml();
        page.setItems(FactsAdaptor.parseSubscribers(xml));
        return page;
    }

    /**
     * Returns a page of block status changes.
     * <p>
     * <a href="http://dev.maileon.com/api/rest-api-1-0/reports/get-blocks/?lang=en">Documentation website</a></p>
     *
     * @param fromDate if provided, only the blocks after the given date will be returned
     * @param toDate if provided, only the blocks before the given date will be returned
     * @param contactIDsFilter multivalued parameter to filter the results by contacts
     * @param contactEmailsFilter multivalued parameter to filter the results by email addresses
     * @param contactExternalIDsFilter multivalued parameter to filter the results by external ids
     * @param reasons multivalued parameter to filter the blocks by reasons. Possible values are: <code>blacklist</code> and <code>bounce_policy</code>.
     * @param oldStatus filter by old status, valid: allowed, blocked
     * @param newStatus Filter by new status, valid: allowed, blocked
     * @param excludeAnonymousBlocks if <code>true</code>, the blocks that cannot be mapped to contacts because of permission types or because of unsubscriptions are omitted in the
     * results.
     * @param standardFields multivalued parameter to query the value of a standard contact field
     * @param customFields multivalued parameter to query the value of a custom contact field
     * @param pageIndex the index of the result page starting from 1
     * @param pageSize the number of {@link Block} in the resulted page
     * @return page containing the corresponding blocks
     * @throws MaileonException
     */
    public Page<Block> getBlocks(Long fromDate, Long toDate, List<Long> contactIDsFilter,
            List<String> contactEmailsFilter, List<String> contactExternalIDsFilter,
            List<String> reasons, String oldStatus, String newStatus,
            boolean excludeAnonymousBlocks, List<StandardContactField> standardFields,
            List<String> customFields, int pageIndex, int pageSize) throws MaileonException {
        validatePaginationParameters(pageIndex, pageSize);
        QueryParameters params = new QueryParameters("page_index", pageIndex);
        params.add("page_size", pageSize);
        params.add("from_date", fromDate);
        params.add("to_date", toDate);
        addParameters(params, standardFields, customFields);
        addParameters(params, contactIDsFilter, contactEmailsFilter, contactExternalIDsFilter);
        params.add("exclude_anonymous_blocks", excludeAnonymousBlocks);
        if (oldStatus != null) {
            params.add("old_status", oldStatus);
        }
        if (newStatus != null) {
            params.add("new_status", newStatus);
        }
        if (reasons != null) {
            params.addList("reasons", new ArrayList<Object>(reasons));
        }
        ResponseWrapper resp = get("reports/blocks", params);
        Page<Block> page = new Page<>();
        page.setPageIndex(pageIndex);
        page.setPageSize(pageSize);
        page.setTotalItems(Integer.parseInt(resp.getHeaders().getFirst("X-Items")));
        page.setCountPages(Integer.parseInt(resp.getHeaders().getFirst("X-Pages")));
        Element xml = resp.getEntityAsXml();
        page.setItems(FactsAdaptor.parseBlocks(xml));
        return page;
    }

    /**
     * Returns the number of blocks.
     *
     * @param fromDate if provided, only the blocks after the given date will be returned
     * @param toDate if provided, only the blocks before the given date will be returned
     * @param contactIDsFilter multivalued parameter to filter the results by contacts
     * @param contactEmailsFilter multivalued parameter to filter the results by email addresses
     * @param contactExternalIDsFilter multivalued parameter to filter the results by external ids
     * @param reasons multivalued parameter to filter the blocks by reasons. Possible values are: <code>blacklist</code> and <code>bounce_policy</code>.
     * @param oldStatus filter by old status, valid: allowed, blocked
     * @param newStatus Filter by new status, valid: allowed, blocked
     * @param excludeAnonymousBlocks if <code>true</code>, the blocks that cannot be mapped to contacts because of permission types or because of unsubscriptions are omitted in the
     * results.
     * @return the count of blocks
     * @throws MaileonException
     */
    public int getBlocksCount(Long fromDate, Long toDate, List<Long> contactIDsFilter,
            List<String> contactEmailsFilter, List<String> contactExternalIDsFilter,
            List<String> reasons, String oldStatus, String newStatus,
            boolean excludeAnonymousBlocks) throws MaileonException {
        QueryParameters params = new QueryParameters();
        params.add("from_date", fromDate);
        params.add("to_date", toDate);
        addParameters(params, contactIDsFilter, contactEmailsFilter, contactExternalIDsFilter);
        params.add("exclude_anonymous_blocks", excludeAnonymousBlocks);
        if (oldStatus != null) {
            params.add("old_status", oldStatus);
        }
        if (newStatus != null) {
            params.add("new_status", newStatus);
        }
        if (reasons != null) {
            params.addList("reasons", new ArrayList<Object>(reasons));
        }
        Element resp = get("reports/blocks/count", params).getEntityAsXml();
        try {
            return Integer.parseInt(resp.getText());
        } catch (Throwable t) {
            throw new MaileonClientException("unexpected response format", t);
        }
    }

    private static void validatePaginationParameters(int pageIndex, int pageSize) throws MaileonBadRequestException {
        if (pageIndex < 1) {
            throw new MaileonBadRequestException("pageIndex must be > 0 - found: " + pageIndex);
        }
        if (pageSize < 1) {
            throw new MaileonBadRequestException("pageSize must be > 0 - found: " + pageSize);
        }
    }

    private static QueryParameters createQueryParameters(int pageIndex, int pageSize, Long fromDate, Long toDate, List<Long> mailingIdFilter) {
        QueryParameters params = new QueryParameters("page_index", pageIndex);
        params.add("page_size", pageSize);
        params.add("from_date", fromDate);
        params.add("to_date", toDate);
        if (mailingIdFilter != null) {
            for (Long id : mailingIdFilter) {
                params.add("mailing_id", id);
            }
        }
        return params;
    }

    private static QueryParameters createQueryParameters(int pageIndex, int pageSize, Long fromDate, Long toDate,
                                                         List<Long> mailingIdFilter, boolean embedFieldBackups) {
        QueryParameters params = createQueryParameters(pageIndex, pageSize, fromDate, toDate, mailingIdFilter);

        params.add("embed_field_backups", embedFieldBackups);
        return params;
    }

    private static void addParameters(QueryParameters params, Format formatFilter, List<String> socialNetworkFilter,
            List<String> deviceTypeFilter, Boolean embedEmailClientInfos) {
        if (formatFilter != null) {
            params.add("format", formatFilter.getValue());
        }
        if (socialNetworkFilter != null) {
            for (String sn : socialNetworkFilter) {
                params.add("social_network", sn);
            }
        }
        if (deviceTypeFilter != null) {
            for (String dt : deviceTypeFilter) {
                params.add("device_type", dt);
            }
        }
        if (embedEmailClientInfos != null) {
            params.add("embed_email_client_infos", embedEmailClientInfos);
        }
    }

    private static void addParameters(QueryParameters params, List<StandardContactField> standardFields, List<String> customFields) {
        params.add(standardFieldParameters(standardFields));
        params.add(customFieldParameters(customFields));
    }

    private static void addParameters(QueryParameters params, List<Long> contactIDsFilter, List<String> contactEmailsFilter, List<String> contactExternalIDsFilter) {
        params.add(createParameters("ids", contactIDsFilter));
        params.add(createParameters("emails", contactEmailsFilter));
        params.add(createParameters("eids", contactExternalIDsFilter));
    }

    /**
     * Adds generic fields with the given name o the parameters
     *
     * @param fieldName The name of the parameter
     * @param fields The values
     * @return the query parameters
     */
    private static <T extends Object> QueryParameters createParameters(String fieldName, List<T> fields) {
        QueryParameters parameters = new QueryParameters();
        if (fields != null) {
            for (Object obj : fields) {
                if (obj != null) {
                    parameters.add(fieldName, obj.toString());
                }
            }
        }
        return parameters;
    }

    /**
     * Standard field parameters.
     *
     * @param standardFields the standard fields
     * @return the query parameters
     */
    private static QueryParameters standardFieldParameters(List<StandardContactField> standardFields) {
        QueryParameters parameters = new QueryParameters();
        if (standardFields != null) {
            for (StandardContactField scf : standardFields) {
                if (scf != null) {
                    parameters.add("standard_field", scf.getName());
                }
            }
        }
        return parameters;
    }

    /**
     * Custom field parameters.
     *
     * @param customFields the custom fields
     * @return the query parameters
     */
    private static QueryParameters customFieldParameters(List<String> customFields) {
        QueryParameters parameters = new QueryParameters();
        if (customFields != null) {
            for (String cf : customFields) {
                if (cf != null) {
                    parameters.add("custom_field", cf);
                }
            }
        }
        return parameters;
    }
}
